import numpy as np
#from TreeExtraction_class import TreeExtraction
#from TreeRepresentation import Jaccard_trees
#from TreeRepresentation import Jaccard_rules
from .TreeRepresentation_utils import tree_splits_to_vector, rule_splits_to_vector


class TreeDissimilarity:
    
    def __init__(self, clf, set_up, tree_indeces, dissim_method= 'rules',
                 feature_represent= 'weighted', sample=None):
        self.clf = clf
        self.set_up = set_up
        self.tree_indeces = tree_indeces
        self.dissim_method = dissim_method
        self.feature_represent = feature_represent
        self.sample = sample #is only used for Jaccard_trees dissimilarity
        
    
    def compute_dissimilarity(self):
        
        vectors = self.tree_to_vectors(self.clf, self.tree_indeces,
                                  self.dissim_method, self.feature_represent,
                                  self.sample)
        
        diss_matrix = self.vectors_to_dissim_matrix(vectors)
        avg_dissimilarity = self.distance_matrix_to_float(diss_matrix)

        return avg_dissimilarity, diss_matrix
    
    
    def tree_to_vectors(self, clf, tree_indeces, dissim_method,
                               feature_represent, sample):
        vectors = [] # list of the vector representation ( tree or path)
        
        if tree_indeces is None: # in case no indeces are given, use all the trees of the ensemble
            tree_indeces = range(clf.n_estimators)
                
        if dissim_method == "trees":
            for idx in tree_indeces:
                # weights "by sample size" or "simple" cases are conisidered
                vectors.append(tree_splits_to_vector(clf, idx, feature_represent))
                
                
        if dissim_method == "rules":
            for idx in tree_indeces:
                # weights "by sample size" or "simple" cases are conisidered
                vectors.append(rule_splits_to_vector(clf, idx, feature_represent,
                                                     sample))    
        return vectors
    
    
    def vectors_to_dissim_matrix(self, vector_list):
        
        size = len(vector_list)
        A = np.zeros([size, size])
        
        for i in range(size):
            for j in range(i, size):
                A[i,j] = np.sum(np.minimum(vector_list[i], vector_list[j]))/np.sum(
                                np.maximum(vector_list[i], vector_list[j]))
                A[j,i] = A[i,j]
        
        return 1-A # dissimilarity instead of similarity (A[i,i]= 0)
    
    
    def distance_matrix_to_float(self, dist_matrix):
        # avergaes the OFF_DIAGONAL elements of the matrix
        # if matrix is (1 x 1) return np.nan (division by zero!)
        if dist_matrix.shape[0] != dist_matrix.shape[1]:
            raise Exception("Non square Matrix?")
        if dist_matrix.shape[0] > 1 and dist_matrix.shape[0] > 1:
            return dist_matrix.sum()/dist_matrix.shape[0]/(dist_matrix.shape[1]-1) 
        else:
            return np.nan
        
    # def Jaccard_trees(self, set_up, clf, idx1, idx2): # it's a SIMILARITY measure
    
    #     the_DTree1 = clf.estimators_[idx1].tree_
    #     the_DTree2 = clf.estimators_[idx2].tree_
        
    #     # in leafs the "feature split" is = -2, and leafs must be dropped
    #     fe1 = np.bincount(the_DTree1.feature[the_DTree1.feature > -1 ]) 
    #     fe2 = np.bincount(the_DTree2.feature[the_DTree2.feature > -1 ])
    #     # PROBLEM: some features might be missing (at the tail), fill with zeros
    #     base1 = np.zeros(clf.n_features_)
    #     base2 = np.zeros(clf.n_features_)
    #     #the bincount will drop all features with no splits that come after the last feature with at least one split
    #     # we need ot inlcude them with extra 0-s at the end to preserve vector lenght. We achieve it by doing the following:
    #     base1[:len(fe1)] += fe1
    #     base2[:len(fe2)] += fe2
    
    #     # Jaccardi formula as of Eq.3 in C443 Suppl. material    
    #     #take "elementwise" min and max along indeces and compute ratio
    
    #     return np.sum(np.minimum(base1, base2))/np.sum(np.maximum(base1, base2))
    
        
    # def trees_to_distance_matrix(self, tree_indeces, *args):
    #                                          # *args: dataset, probabs y/n
    #     size = len(tree_indeces)
    #     M = np.zeros([size, size])
    #     for i in range(size):
    #         for j in range(i+1):
    #             if self.dissim_method == "Jaccard":
    #                 M[i,j] = 1-self.Jaccard_index(i, j)
    #                         #1-Jacc for dissimilarity
    #             else:
    #                 raise NameError('input key not recognised, typo?')
    #             M[j,i] = M[i,j]
    #     return M
    
    #def tree_to_vector_representation(self, idx):
        
    # def Jaccard_index(self, idx1, idx2): # it's a SIMILARITY measure
        
    #     # we compute Jaccard similarity index given two tree (indeces)
    #     # of the original R(S)F ensemble learner

    #     the_DTree1 = self.clf.estimators_[idx1].tree_
    #     the_DTree2 = self.clf.estimators_[idx2].tree_
        
    #     # in leafs the "feature split" is = -2, and leafs must be dropped
    #     split_freq1 = np.bincount(the_DTree1.feature[the_DTree1.feature > -1 ]) 
    #     split_freq2 = np.bincount(the_DTree2.feature[the_DTree2.feature > -1 ])
    #     # PROBLEM: some features might be missing (at the tail), fill with zeros
    #     vec1 = np.zeros(self.clf.n_features_)
    #     vec2 = np.zeros(self.clf.n_features_)
    #     #the bincount will drop all features with no splits that come after the last feature with at least one split
    #     # we need ot inlcude them with extra 0-s at the end to preserve vector lenght.
    #     #We achieve it by doing the following:
    #     vec1[:len(split_freq1)] += split_freq1
    #     vec2[:len(split_freq2)] += split_freq2
    
    #     # Jaccard formula as of Eq.3 in C443 Suppl. materil
    #     #take "elementwise" min and max along indeces and compute ratio
    #     return np.sum(np.minimum(vec1, vec2))/np.sum(np.maximum(vec1, vec2))
    #             #Jaccardi sklearn should correspond

    


        


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 22 11:21:19 2024

@author: Klest Dedja
"""
import os #TODO: consider moving to pathlib for cross-OS consistency.
from setuptools import find_packages, setup

readme_path = os.path.join('README.md')

with open(readme_path, 'r') as f:
    text_description = f.read()


version_file_path = os.path.join('app', 'bellatrex', 'version.txt')

with open(version_file_path) as version_file:
    synch_version = version_file.read().strip()

setup(
    name='bellatrex',
    version=synch_version,
    description='A toolbox for Building Explanations through a LocaLly AccuraTe Rule EXtractor',
    package_dir={'': 'app'}, #where to find the code: in the app subfolder
    packages=find_packages(where='app'),
    include_package_data=True, #include additional files (which ones precisely is determined by the MANIFEST.in file)
    package_data={
        'bellatrex.datasets': ['*.csv'],
    },
    long_description=text_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Klest94/Bellatrex',
    author='Klest Dedja',
    author_email='daneel.olivaw94@gmail.com',
    licence='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',    ],
    install_requires=['scikit-learn >= 1.2',
                      'threadpoolctl>=3.1',
                      'scikit-survival>=0.22',
                      'scipy>=1.11',
                      'pandas>=1.5', #to load tutorial datasets
                      'matplotlib>=3.7'],
    extras_require={
        'dev': ['pytest', 'twine'],
        'gui': ['dearpygui>=1.6.2', 'dearpygui-ext>=0.9.5'] # for Graphical User Interface (demo)
    },
    python_requires='>=3.9, <4',
)



import os
import matplotlib.pyplot as plt

import bellatrex as btrex
print("Bellatrex version:", btrex.__version__)

MAX_TEST_SAMPLES = 2
PLOT_GUI = False

root_folder = os.getcwd()
print(root_folder)

from sksurv.ensemble import RandomSurvivalForest
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor

from sklearn.model_selection import train_test_split


from bellatrex.datasets import load_mtr_data, load_mlc_data
from bellatrex.datasets import load_survival_data, load_binary_data, load_regression_data
from bellatrex.utilities import get_auto_setup

# X, y = load_binary_data(return_X_y=True)
X, y = load_regression_data(return_X_y=True)
# X, y = load_survival_data(return_X_y=True)
# X, y = load_mlc_data(return_X_y=True)
# X, y = load_mtr_data(return_X_y=True)

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=0)

SETUP = get_auto_setup(y)
print('Detected prediction task \'SETUP\':', SETUP)


### instantiate original R(S)F estimator, works best with some pruning.
if SETUP.lower() in 'survival':
    clf = RandomSurvivalForest(n_estimators=100, min_samples_split=10,
                                n_jobs=-2, random_state=0)

elif SETUP.lower() in ['binary', 'multi-label']:
    clf = RandomForestClassifier(n_estimators=100, min_samples_split=5,
                                n_jobs=-2, random_state=0)

elif SETUP.lower() in ['regression', 'multi-target']:
    clf = RandomForestRegressor(n_estimators=100, min_samples_split=5,
                                n_jobs=-2, random_state=0)


from bellatrex import BellatrexExplain
from bellatrex.wrapper_class import EnsembleWrapper, pack_trained_ensemble

clf.fit(X_train, y_train)
print('Model fitting complete.')

clf_packed = pack_trained_ensemble(clf)

# Load the pretrained model and make it compatible with Bellatrex through EnsembleWrapper()
clf2 = EnsembleWrapper(clf_packed)


#fit RF here. The hyperparameters are given
# compatible with trained model clf, and with a wrapped dictionary as in clf1
Btrex_fitted = BellatrexExplain(clf2, set_up='auto',
                                p_grid={"n_clusters": [1, 2, 3]},
                                verbose=3).fit(X_train, y_train)


for i in range(MAX_TEST_SAMPLES): # iterate for the first few samples in the test set

    print(f"Explaining sample i={i}")

    # Store the extracted rules in dedicated folder, if needed:
    explan_dir = os.path.join(root_folder, "explanations-out")
    os.makedirs(explan_dir, exist_ok=True)
    FILE_OUT = os.path.join(explan_dir, "Rules_"+str(SETUP)+'_id'+str(i)+'.txt')

    # y_train_pred = clf.predict_proba(X_train)[:,1]
    # y_test_pred = clf.predict_proba(X_test)[:,1]
    # y_train_pred = clf.predict(X_train)
    # y_test_pred = clf.predict(X_test)

    tuned_method = Btrex_fitted.explain(X_test, i)
    tuned_method.plot_overview(show=True)

    tuned_method.plot_visuals(plot_max_depth=5,
                              preds_distr=None,
                              conf_level=0.9,
                              tot_digits=4)
    plt.show()

    # tuned_method.plot_overview(show=True)

# if using Qt5Agg backend, the figures will be deleted unless we call the following:
# plt.show()

# %%
